/*--------------------------------------------------------------------------+
$Id: codetemplates.xml 1 2011-01-01 00:00:01Z hoelzl $
|                                                                          |
| Copyright 2016 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.internal;

import static org.conqat.lib.commons.collections.CollectionUtils.asUnmodifiable;

import java.util.ArrayList;
import java.util.Collection;
import java.util.EventObject;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.conqat.lib.commons.collections.UnmodifiableMap;
import org.eclipse.emf.common.command.CommandStackListener;
import org.eclipse.emf.ecore.EObject;
import org.fortiss.tooling.kernel.extension.ITutorialProvider;
import org.fortiss.tooling.kernel.extension.ITutorialStepCompletionChecker;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.extension.data.Prototype;
import org.fortiss.tooling.kernel.extension.data.TutorialStepAtomic;
import org.fortiss.tooling.kernel.extension.data.TutorialStepBase;
import org.fortiss.tooling.kernel.extension.data.TutorialStepComposite;
import org.fortiss.tooling.kernel.service.ICommandStackService;
import org.fortiss.tooling.kernel.service.IPersistencyService;
import org.fortiss.tooling.kernel.service.ITutorialService;
import org.fortiss.tooling.kernel.service.listener.ITutorialServiceListener;

/**
 * This class implements the {@link ITutorialService} interface.
 * 
 * @author hoelzl
 * @author $Author: hoelzl $
 * @version $Rev: 18709 $
 * @ConQAT.Rating RED Hash:
 */
public final class TutorialService implements ITutorialService, CommandStackListener {
	/** The collection of all defined tutorials. */
	private Map<String, Map<String, Class<? extends ITutorialProvider>>> allTutorials =
			new HashMap<>();
	/** The collection of listeners. */
	private Collection<ITutorialServiceListener> listeners =
			new LinkedList<ITutorialServiceListener>();
	/** The list of atomic steps. */
	private List<TutorialStepAtomic> atomicStepList = new ArrayList<TutorialStepAtomic>();

	/** The active tutorial. */
	private ITutorialProvider activeTutorial = null;
	/** The active tutorial step. */
	private TutorialStepAtomic activeStep = null;
	/** The active top-level element. */
	private ITopLevelElement activeTopElement = null;

	/** {@inheritDoc} */
	@Override
	public boolean isTutorialActive() {
		return activeTutorial != null;
	}

	/** {@inheritDoc} */
	@Override
	public ITutorialProvider getActiveTutorial() {
		return activeTutorial;
	}

	/** {@inheritDoc} */
	@Override
	public TutorialStepBase getActiveTutorialStep() {
		return activeStep;
	}

	/** {@inheritDoc} */
	@Override
	public Map<String, Map<String, Class<? extends ITutorialProvider>>> getAllTutorialProviders() {
		return new UnmodifiableMap<String, Map<String, Class<? extends ITutorialProvider>>>(
				allTutorials);
	}

	/** {@inheritDoc} */
	@Override
	public void registerTutorialProvider(Class<? extends ITutorialProvider> provider,
			String displayName, String category) {
		if(category == null) {
			category = "";
		}
		Map<String, Class<? extends ITutorialProvider>> categoryMap = allTutorials.get(category);
		if(categoryMap == null) {
			categoryMap = new HashMap<String, Class<? extends ITutorialProvider>>();
			allTutorials.put(category, categoryMap);
		}
		categoryMap.put(displayName, provider);
	}

	/** {@inheritDoc} */
	@Override
	public void startTutorial(Class<? extends ITutorialProvider> provider) {
		try {
			activeTutorial = provider.newInstance();
			List<TutorialStepBase> rootSteps = activeTutorial.getDefinition().getSteps();
			for(TutorialStepBase step : rootSteps) {
				collectAtomicSteps(step);
			}
			activeStep = atomicStepList.get(0);
			EObject root = activeTutorial.getRootElement();
			activeTopElement = IPersistencyService.INSTANCE.addDummyEObjectAsTopLevelElement(root);
			ICommandStackService.INSTANCE.addCommandStackListener(activeTopElement, this);
		} catch(InstantiationException e) {
			e.printStackTrace();
		} catch(IllegalAccessException e) {
			e.printStackTrace();
		}
	}

	/** Collects all atomic steps. */
	private void collectAtomicSteps(TutorialStepBase step) {
		if(step instanceof TutorialStepAtomic) {
			atomicStepList.add((TutorialStepAtomic)step);
		} else if(step instanceof TutorialStepComposite) {
			TutorialStepComposite tsc = (TutorialStepComposite)step;
			for(TutorialStepBase s : tsc.getSteps()) {
				collectAtomicSteps(s);
			}
		}
	}

	/** {@inheritDoc} */
	@Override
	public void stopTutorial() {
		ICommandStackService.INSTANCE.removeCommandStackListener(activeTopElement, this);
		IPersistencyService.INSTANCE.removeDummyTopLevelElement(activeTutorial.getRootElement());
		activeTopElement = null;
		atomicStepList.clear();
		activeStep = null;
		activeTutorial = null;
	}

	/** {@inheritDoc} */
	@Override
	public void addTutorialServiceListener(ITutorialServiceListener l) {
		if(!listeners.contains(l)) {
			listeners.add(l);
		}
	}

	/** {@inheritDoc} */
	@Override
	public void removeTutorialServiceListener(ITutorialServiceListener l) {
		listeners.remove(l);
	}

	/** {@inheritDoc} */
	@Override
	public boolean prototypeActive(Prototype prototype) {
		return activeTutorial.getGlobalWhitelistProvider().prototypeActive(prototype) ||
				activeStep.getLocalWhitelistProvider().prototypeActive(prototype);
	}

	/** {@inheritDoc} */
	@Override
	public boolean isEarlierStep(TutorialStepBase step) {
		if(activeStep == TUTORIAL_COMPLETED_STEP) {
			return true;
		}
		if(step instanceof TutorialStepComposite) {
			TutorialStepComposite cmp = (TutorialStepComposite)step;
			TutorialStepBase inner = cmp.getSteps().get(cmp.getSteps().size() - 1);
			return isEarlierStep(inner);
		}
		// TutorialStepAtomic
		int current = atomicStepList.indexOf(activeStep);
		int test = atomicStepList.indexOf(step);
		return test < current;
	}

	/** {@inheritDoc} */
	@Override
	public boolean isLaterStep(TutorialStepBase step) {
		if(activeStep == TUTORIAL_COMPLETED_STEP) {
			return false;
		}
		if(step instanceof TutorialStepComposite) {
			TutorialStepComposite cmp = (TutorialStepComposite)step;
			TutorialStepBase inner = cmp.getSteps().get(cmp.getSteps().size() - 1);
			return isLaterStep(inner);
		}
		// TutorialStepAtomic
		int current = atomicStepList.indexOf(activeStep);
		int test = atomicStepList.indexOf(step);
		return test > current;
	}

	/** {@inheritDoc} */
	@Override
	public void commandStackChanged(EventObject event) {
		activeStep = TUTORIAL_COMPLETED_STEP;
		for(TutorialStepAtomic step : atomicStepList) {
			if(!step.getCompletionChecker().isTutorialStepComplete(activeTutorial.getRootElement())) {
				activeStep = step;
				break;
			}
		}
	}

	/** The shared hidden final state of all tutorials. */
	private static TutorialStepAtomic TUTORIAL_COMPLETED_STEP = new TutorialStepAtomic(
			"Tutorial complete.") {
		@Override
		public ITutorialStepCompletionChecker getCompletionChecker() {
			return new ITutorialStepCompletionChecker() {
				@Override
				public boolean isTutorialStepComplete(EObject model) {
					// This step can never be completed.
					return false;
				}
			};
		}
	};

	/** {@inheritDoc} */
	@Override
	public List<TutorialStepAtomic> getAtomicTutorialSteps() {
		return asUnmodifiable(atomicStepList);
	}
}
